from .connection import Connection
from .document import Document, EmbeddedDocument
from .fields import (ObjectIdField, IntField, BooleanField, FloatField,
                     BinaryField, StringField, EmailField, DateTimeField,
                     DictField, ListField, EmbeddedField, SequenceField,
                     AnyField)
from .metatype import ShardKey, IDFormatter, Index

__all__ = ['Connection', 'Document', 'EmbeddedDocument', 'AnyField',
           'ObjectIdField', 'IntField', 'BooleanField', 'FloatField',
           'BinaryField', 'StringField', 'EmailField', 'DateTimeField',
           'DictField', 'ListField', 'EmbeddedField', 'SequenceField',
           'ShardKey', 'IDFormatter', 'Index']

__version__ = '0.2.16'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
from enum import Enum
from datetime import datetime

from bson import ObjectId

from .errors import ValidationError, DeserializationError, ArgumentError


# MongoDB will not store dates with milliseconds.
def milli_trim(x):
    return x.replace(
        microsecond=int((x.microsecond / 1000) * 1000))


class BaseField(object):

    """ Base field for all fields. """
    types = []

    def __init__(self, default=None, required=False,
                 nullable=False, name=None):
        self.default = default
        self.required = required
        self.nullable = nullable

        # if None, will be set to assigned name in metaclass
        self.name = name

    def _raise_validation_error(self, value):
        raise ValidationError(self.__class__, self.name, value)

    def validate(self, value):
        if self.required and not self.nullable:
            if value is None:
                self._raise_validation_error(value)
            elif self.types:
                for type_ in self.types:
                    if isinstance(value, type_):
                        break
                else:
                    self._raise_validation_error(value)

    def pre_save_val(self, value):
        """ do something to field value before save """
        return None

    def to_storage(self, value):
        return value

    def to_python(self, value):
        return value


AnyField = BaseField


class ObjectIdField(BaseField):
    types = [ObjectId]


class IntField(BaseField):
    types = [int]

    def __init__(self, min=None, max=None, **kwargs):
        self._min = min
        self._max = max
        super(IntField, self).__init__(**kwargs)

    def validate(self, value):
        super(IntField, self).validate(value)

        if value:
            if (self._min and self._min > value) or \
                    (self._max and self._max < value):
                self._raise_validation_error(value)


class EnumField(BaseField):
    types = [Enum]

    def __init__(self, cls, **kwargs):
        self._cls = cls
        super(EnumField, self).__init__(**kwargs)

    def validate(self, value):
        super(EnumField, self).validate(value)

        if value:
            if not isinstance(value, self._cls):
                self._raise_validation_error(value)


class BooleanField(BaseField):
    types = [bool]


class FloatField(BaseField):
    types = [float, int]


class BinaryField(BaseField):
    types = [bytes]

    def __init__(self, min_bytes=None, max_bytes=None, **kwargs):
        self.min_bytes = min_bytes
        self.max_bytes = max_bytes
        super(BinaryField, self).__init__(**kwargs)

    def validate(self, value):
        super(BinaryField, self).validate(value)
        if value:
            if (self.min_bytes and len(value) < self.min_bytes) or \
                    (self.max_bytes and len(value) > self.max_bytes):
                self._raise_validation_error(value)


class StringField(BaseField):
    types = [str]

    def __init__(self, min_length=None, max_length=None, strip=True, **kwargs):
        self.min_length = min_length
        self.max_length = max_length
        self.strip = strip
        super(StringField, self).__init__(**kwargs)

    def validate(self, value):
        super(StringField, self).validate(value)
        if value:
            if (self.min_length and len(value) < self.min_length) or \
                    (self.max_length and len(value) > self.max_length):
                self._raise_validation_error(value)

    def to_storage(self, value):
        if self.strip and value:
            value = value.strip()
        return value


class EmailField(StringField):
    email_re = re.compile(r'^[\S]+@[\S]+\.[\S]+$')

    def __init__(self, **kwargs):
        super(EmailField, self).__init__(
            min_length=5, max_length=100, **kwargs)

    def validate(self, value):
        super(EmailField, self).validate(value)

        if value and not self.email_re.match(value):
            self._raise_validation_error(value)


class DateTimeField(BaseField):
    types = [datetime]

    def __init__(self, modified=False, created=False, **kwargs):

        self.modified = modified
        self.created = created
        if self.created:
            def _default():
                return datetime.utcnow()
            default = _default
        else:
            default = None

        super(DateTimeField, self).__init__(default=default, **kwargs)

    def pre_save_val(self, value):
        if self.modified:
            return datetime.utcnow()

    def to_python(self, value):
        if value is None or isinstance(value, datetime):
            return value

        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

    def to_storage(self, value):
        if value and isinstance(value, datetime):
            return milli_trim(value)
        return value


class DictField(BaseField):
    types = [dict]

    def __init__(self, default=None, **kwargs):
        default = default or {}
        super(DictField, self).__init__(default=default, **kwargs)

    def to_python(self, value):
        if value is None:
            return {}

        if not isinstance(value, dict):
            raise DeserializationError(self, value)

        return value


class ListField(BaseField):
    types = [list]

    def __init__(self, field=None, default=None, **kwargs):
        default = default or []
        self.field = field
        super(ListField, self).__init__(default=default, **kwargs)

    def validate(self, value):
        super(ListField, self).validate(value)

        if value and self.field:
            for v in value:
                self.field.validate(v)

    def to_storage(self, value):
        if self.field:
            return [self.field.to_storage(v) for v in value]
        else:
            return value

    def to_python(self, value):
        if value is None:
            return []

        if not isinstance(value, list):
            raise DeserializationError(self, value)

        if self.field:
            return [self.field.to_python(v) for v in value]
        else:
            return value


class EmbeddedField(BaseField):

    def __init__(self, embedded=None, **kwargs):
        from yamo import EmbeddedDocument
        if not issubclass(embedded, EmbeddedDocument):
            raise ArgumentError(EmbeddedField, embedded)

        super(EmbeddedField, self).__init__(**kwargs)
        self.embedded = embedded

    def validate(self, value):
        super(EmbeddedField, self).validate(value)

        if value:
            if not isinstance(value, self.embedded):
                self._raise_validation_error(value)
            else:
                value.validate()

    def to_storage(self, value):
        if isinstance(value, dict):
            value = self.embedded(value)

        return getattr(value, '_data', None)

    def to_python(self, value):
        if value:
            return self.embedded(value)


class SequenceField(IntField):

    """ Auto Increment Integer Field """

    def pre_save_val(self, value):
        if value:
            return value

        while True:
            try:
                r = self._doc._db.counters.find_and_modify(
                    query={'_id': self.name},
                    update={'$inc': {'seq': 1}},
                    new=True, upsert=True)
            except:
                continue
            else:
                break
        return r['seq']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import types
import pickle
import functools


from pymongo.cursor import Cursor


class CachedModel(object):

    """ Used in Model.cached

    :param timeout: timeout in seconds
    :param cache_none: whether to cache None results
    """
    # cls -> (cache_time, value)
    caches = {}

    def __init__(self, cls, timeout=300, cache_none=False):
        self.cls = cls
        self.timeout = timeout
        self.cache_none = cache_none
        self.count = 0
        if self.cls not in self.caches:
            self.caches[self.cls] = {}

    def _clear_timeout(self):
        cache = self.caches[self.cls]
        now = time.time()
        invals = []
        for key, values in cache.items():
            if values[0] < now - self.timeout:
                invals.append(key)
        for key in invals:
            del cache[key]

    def __getattr__(self, name):
        self.count += 1
        if self.count % 1000 == 0:
            self._clear_timeout()

        attr = getattr(self.cls, name)
        if callable(attr):
            # wrap this callable to use cache
            @functools.wraps(attr)
            def deco(*args, **kwargs):
                cache = self.caches[self.cls]
                key = pickle.dumps([attr.__name__, args, kwargs])
                cache_miss = key not in cache

                def timedout():
                    return cache[key][0] < time.time() - self.timeout

                if cache_miss or timedout():
                    value = attr(*args, **kwargs)
                    if isinstance(value, Cursor) or \
                            isinstance(value, types.GeneratorType):
                        # this will consume A LOT of memory, use with care
                        value = list(value)
                    if value is not None or self.cache_none:
                        cache[key] = (time.time(), value)
                    else:
                        return
                return cache[key][1]
            return deco
        else:
            return attr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from collections import OrderedDict

from .fields import BaseField
from .errors import ArgumentError
from .connection import Connection

log = logging.getLogger('yamo')


class EmbeddedDocumentType(type):

    """
    This is a type that generates EmbeddedDocument classes properly,
    setting their attributes not to be instances of Field, but rather an
    API for MongoDB itself.
    """

    @staticmethod
    def get_maker(attr):
        def getter(self, attr=attr):
            return self._fields[attr].to_python(self._data.get(attr))

        return getter

    @staticmethod
    def set_maker(attr):
        def setter(self, val=None, attr=attr):
            self._data[attr] = self._fields[attr].to_storage(val)

        return setter

    def __new__(cls, name, bases, dct):
        dct.setdefault('_fields', {})

        for x in bases:
            dct['_fields'].update(getattr(x, '_fields', {}))

        for attr, val in dct.items():
            if isinstance(val, BaseField):
                val.name = attr
                dct['_fields'][attr] = val
                dct[attr] = property(cls.get_maker(attr),
                                     cls.set_maker(attr))

        new_cls = super(EmbeddedDocumentType, cls).__new__(
            cls, name, bases, dct)
        return new_cls


class DocumentType(EmbeddedDocumentType):

    """
    This is a type that generates Document classes properly, setting their
    attributes not to be instances of Field, but rather an API for MongoDB
    itself.
    """
    def __new__(cls, name, bases, dct):
        def build_meta(val):
            val._indexes = []
            val._shardkey = None
            val._formatter = None
            for k, v in val.__dict__.items():
                if k not in ['__weakref__', '__doc__',
                             '__dict__', '__module__']:
                    if isinstance(v, Index):
                        val._indexes.append(v)
                    elif isinstance(v, ShardKey):
                        val._shardkey = v
                    elif isinstance(v, IDFormatter):
                        val._formatter = v

            if not hasattr(val, '__collection__'):
                setattr(val, '__collection__', name.lower())

        if '_id' not in dct:
            def idgetter(self):
                return self._data.get('_id')

            def idsetter(self, val):
                self._data['_id'] = val

            dct['_id'] = property(idgetter, idsetter)

        if 'Meta' not in dct:
            class Meta:
                pass
            dct['Meta'] = Meta
        elif not type(dct['Meta']) is type:
            raise ArgumentError(name, dct['Meta'])

        build_meta(dct['Meta'])

        new_cls = super(DocumentType, cls).__new__(cls, name, bases, dct)

        # setting up connection hook
        Connection.docdb[new_cls] = None
        return new_cls


class Index(object):

    """ Index for Document

    >>> class Post(Document):
    ...    class Meta:
    ...        idx1 = Index('created_at')
    ...        Index(['author', 'tag'], sparse=True)
    ...    pass
    """

    def __init__(self, keys, **kwargs):
        """ MongoDB Index

        Takes either a single key or a list of (key, direction) pairs.
        The key(s) must be an instance of :class:`basestring`
        (:class:`str` in python 3), and the direction(s) must be one of
        (:data:`~pymongo.ASCENDING`, :data:`~pymongo.DESCENDING`,
        :data:`~pymongo.GEO2D`, :data:`~pymongo.GEOHAYSTACK`,
        :data:`~pymongo.GEOSPHERE`, :data:`~pymongo.HASHED`,
        :data:`~pymongo.TEXT`)

        Valid kwargs include, but are not limited to:

          - `name`: custom name to use for this index - if none is
            given, a name will be generated.
          - `unique`: if ``True`` creates a uniqueness constraint on the index.
          - `background`: if ``True`` this index should be created in the
            background.
          - `sparse`: if ``True``, omit from the index any documents that lack
            the indexed field.
          - `bucketSize`: for use with geoHaystack indexes.
            Number of documents to group together within a certain proximity
            to a given longitude and latitude.
          - `min`: minimum value for keys in a :data:`~pymongo.GEO2D`
            index.
          - `max`: maximum value for keys in a :data:`~pymongo.GEO2D`
            index.
          - `expireAfterSeconds`: <int> Used to create an expiring (TTL)
            collection. MongoDB will automatically delete documents from
            this collection after <int> seconds. The indexed field must
            be a UTC datetime or the data will not expire.

        see pymongo's create_index function
        """
        if isinstance(keys, str):
            keys = [keys]
        self.keys = []
        for key in keys:
            if isinstance(key, str):
                self.keys.append((key, 1))
            elif isinstance(key, tuple) and len(key) == 2:
                self.keys.append(key)
            else:
                raise ArgumentError(Index, keys)
        self.kwargs = kwargs
        self.kwargs['background'] = True


class ShardKey(object):

    """ ShardKey for Document

    >>> class Post(Document):
    ...     class Meta:
    ...         ShardKey([('author', 'hash')])

    """

    def __init__(self, keys):
        if isinstance(keys, str):
            keys = [keys]

        self.key = OrderedDict()
        for key in keys:
            if isinstance(key, str):
                key = (key, 1)
            self.key[key[0]] = key[1]


class IDFormatter(object):

    """ IDFormatter for Document

    >>> class Post(Document):
    ...     class Meta:
    ...         IDFormatter('{author}_{pid}')

    can also use a callable function
    """

    def __init__(self, tmpl_or_cb):
        if callable(tmpl_or_cb):
            self._format = tmpl_or_cb
        elif isinstance(tmpl_or_cb, str):
            def _format(**kwargs):
                return tmpl_or_cb.format(**kwargs)
            self._format = _format
        else:
            raise ArgumentError(IDFormatter, tmpl_or_cb)
